#include "clist_patch.h"

#include <stdio.h> /* DEBUG */

/* ----------------------------------------------------------------------- */
/* code copied from gtkclist.c (gtk+ release 1.2.8 ) --------------------- */
/* ----------------------------------------------------------------------- */

#define COLUMN_INSET 3
#define CELL_SPACING 1
#define DRAG_WIDTH  6
#define CLIST_OPTIMUM_SIZE 10

#define GTK_CLIST_CLASS_FW(_widget_) GTK_CLIST_CLASS (((GtkObject*) (_widget_))->klass)

#define min(a,b) ((a) < (b) ? (a) : (b))

#define CLIST_UNFROZEN(clist)     (((GtkCList*) (clist))->freeze_count == 0)
#define	CLIST_REFRESH(clist)	G_STMT_START { \
  if (CLIST_UNFROZEN (clist)) \
    GTK_CLIST_CLASS_FW (clist)->refresh ((GtkCList*) (clist)); \
} G_STMT_END

static inline gint
LIST_WIDTH (GtkCList * clist) 
{
  gint last_column;

  for (last_column = clist->columns - 1;
       last_column >= 0 && !clist->column[last_column].visible; last_column--);

  if (last_column >= 0)
    return (clist->column[last_column].area.x +
	    clist->column[last_column].area.width +
	    COLUMN_INSET + CELL_SPACING);
  return 0;
}

static void
size_allocate_title_buttons (GtkCList *clist)
{
  GtkAllocation button_allocation;
  gint last_column;
  gint last_button = 0;
  gint i;

  if (!GTK_WIDGET_REALIZED (clist))
    return;

  button_allocation.x = clist->hoffset;
  button_allocation.y = 0;
  button_allocation.width = 0;
  button_allocation.height = clist->column_title_area.height;

  /* find last visible column */
  for (last_column = clist->columns - 1; last_column >= 0; last_column--)
    if (clist->column[last_column].visible)
      break;

  for (i = 0; i < last_column; i++)
    {
      if (!clist->column[i].visible)
	{
	  last_button = i + 1;
	  gdk_window_hide (clist->column[i].window);
	  continue;
	}

      button_allocation.width += (clist->column[i].area.width +
				  CELL_SPACING + 2 * COLUMN_INSET);

      if (!clist->column[i + 1].button)
	{
	  gdk_window_hide (clist->column[i].window);
	  continue;
	}

      gtk_widget_size_allocate (clist->column[last_button].button,
				&button_allocation);
      button_allocation.x += button_allocation.width;
      button_allocation.width = 0;

      if (clist->column[last_button].resizeable)
	{
	  gdk_window_show (clist->column[last_button].window);
	  gdk_window_move_resize (clist->column[last_button].window,
				  button_allocation.x - (DRAG_WIDTH / 2), 
				  0, DRAG_WIDTH,
				  clist->column_title_area.height);
	}
      else
	gdk_window_hide (clist->column[last_button].window);

      last_button = i + 1;
    }

  button_allocation.width += (clist->column[last_column].area.width +
			      2 * (CELL_SPACING + COLUMN_INSET));
  gtk_widget_size_allocate (clist->column[last_button].button,
			    &button_allocation);

  if (clist->column[last_button].resizeable)
    {
      button_allocation.x += button_allocation.width;

      gdk_window_show (clist->column[last_button].window);
      gdk_window_move_resize (clist->column[last_button].window,
			      button_allocation.x - (DRAG_WIDTH / 2), 
			      0, DRAG_WIDTH, clist->column_title_area.height);
    }
  else
    gdk_window_hide (clist->column[last_button].window);
}

static void
size_allocate_columns (GtkCList *clist,
		       gboolean  block_resize)
{
  gint xoffset = CELL_SPACING + COLUMN_INSET;
  gint last_column;
  gint i;

  /* find last visible column and calculate correct column width */
  for (last_column = clist->columns - 1;
       last_column >= 0 && !clist->column[last_column].visible; last_column--);

  if (last_column < 0)
    return;

  for (i = 0; i <= last_column; i++)
    {
      if (!clist->column[i].visible)
	continue;
      clist->column[i].area.x = xoffset;
      if (clist->column[i].width_set)
	{
	  if (!block_resize && GTK_CLIST_SHOW_TITLES(clist) &&
	      clist->column[i].auto_resize && clist->column[i].button)
	    {
	      gint width;

	      width = (clist->column[i].button->requisition.width -
		       (CELL_SPACING + (2 * COLUMN_INSET)));

	      if (width > clist->column[i].width)
		gtk_clist_set_column_width (clist, i, width);
	    }

	  clist->column[i].area.width = clist->column[i].width;
	  xoffset += clist->column[i].width + CELL_SPACING + (2* COLUMN_INSET);
	}
      else if (GTK_CLIST_SHOW_TITLES(clist) && clist->column[i].button)
	{
	  clist->column[i].area.width =
	    clist->column[i].button->requisition.width -
	    (CELL_SPACING + (2 * COLUMN_INSET));
	  xoffset += clist->column[i].button->requisition.width;
	}
    }

  clist->column[last_column].area.width = clist->column[last_column].area.width
    + MAX (0, clist->clist_window_width + COLUMN_INSET - xoffset);
}

/*

struct _GtkCListColumn
{
  gchar *title;
  GdkRectangle area;
  
  GtkWidget *button;
  GdkWindow *window;
  
  gint width;
  gint min_width;
  gint max_width;
  GtkJustification justification;
  
  guint visible        : 1;  
  guint width_set      : 1;
  guint resizeable     : 1;
  guint auto_resize    : 1;
  guint button_passive : 1;
};

struct _GtkCListRow
{
  GtkCell *cell;
  GtkStateType state;
  
  GdkColor foreground;
  GdkColor background;
  
  GtkStyle *style;

  gpointer data;
  GtkDestroyNotify destroy;
  
  guint fg_set     : 1;
  guint bg_set     : 1;
  guint selectable : 1;
};

--- Cell Structures ---

struct _GtkCellText
{
  GtkCellType type;
  
  gint16 vertical;
  gint16 horizontal;
  
  GtkStyle *style;

  gchar *text;
};

struct _GtkCellPixmap
{
  GtkCellType type;
  
  gint16 vertical;
  gint16 horizontal;
  
  GtkStyle *style;

  GdkPixmap *pixmap;
  GdkBitmap *mask;
};

struct _GtkCellPixText
{
  GtkCellType type;
  
  gint16 vertical;
  gint16 horizontal;
  
  GtkStyle *style;

  gchar *text;
  guint8 spacing;
  GdkPixmap *pixmap;
  GdkBitmap *mask;
};

struct _GtkCellWidget
{
  GtkCellType type;
  
  gint16 vertical;
  gint16 horizontal;
  
  GtkStyle *style;

  GtkWidget *widget;
};

struct _GtkCell
{
  GtkCellType type;
  
  gint16 vertical;
  gint16 horizontal;
  
  GtkStyle *style;

  union {
    gchar *text;
    
    struct {
      GdkPixmap *pixmap;
      GdkBitmap *mask;
    } pm;
    
    struct {
      gchar *text;
      guint8 spacing;
      GdkPixmap *pixmap;
      GdkBitmap *mask;
    } pt;
    
    GtkWidget *widget;
  } u;
};

 */

/* ----------------------------------------------------------------------- */
/* new function: gtk_clist_set_columns ------------------------------- */
/* ----------------------------------------------------------------------- */

static void copy_cell (GtkCell *t, GtkCell *f)
{
	t->type = f->type;
	t->vertical = f->vertical;
	t->horizontal = f->horizontal;
	t->style = f->style;

	switch (t->type)
	{
	case GTK_CELL_TEXT:
		t->u.text = f->u.text;

//		f->u.text = 0;
		break;

	case GTK_CELL_PIXMAP:
		t->u.pm.pixmap = f->u.pm.pixmap;
		t->u.pm.mask = f->u.pm.mask;

//		f->u.pm.pixmap = 0;
//		f->u.pm.mask = 0;
		break;

	case GTK_CELL_PIXTEXT:
		t->u.pt.text = f->u.pt.text;
		t->u.pt.spacing = f->u.pt.spacing;
		t->u.pt.pixmap = f->u.pt.pixmap;
		t->u.pt.mask = f->u.pt.mask;

//		f->u.pt.text = 0;
//		f->u.pt.spacing = 0;
//		f->u.pt.pixmap = 0;
//		f->u.pt.mask = 0;
		break;

	case GTK_CELL_WIDGET:
		t->u.widget = f->u.widget;

//		f->u.widget = 0;
		break;

	default: // _EMPTY, etc...
		break;
	}

//	f->type = GTK_CELL_EMPTY;
}

void gtk_clist_set_columns (GtkCList *list, int ncols)
{
	GtkCListColumn *col;
	GMemChunk *mchunk;
	int c, l, ocols;
	GList *ri;
	GtkCListRow *r;
	GdkWindowAttr attributes;
	gint attributes_mask;
	GtkCell *cell;

	g_return_if_fail(list != NULL);
	g_return_if_fail(ncols > 0);

	if (list->columns == ncols) return;

	col = g_new(GtkCListColumn,ncols);

	ocols = list->columns;
	l = min(ncols,ocols);

	for (c = 0; c < l; ++c)
	{
		col[c].title = list->column[c].title;
		col[c].area.x = list->column[c].area.x;
		col[c].area.y = list->column[c].area.y;
		col[c].area.width = list->column[c].area.width;
		col[c].area.height = list->column[c].area.height;
		col[c].button = list->column[c].button;
		col[c].window = list->column[c].window;
		col[c].width = list->column[c].width;
		col[c].min_width = list->column[c].min_width;
		col[c].max_width = list->column[c].max_width;
		col[c].justification = list->column[c].justification;
		col[c].visible = list->column[c].visible;
		col[c].width_set = list->column[c].width_set;
		col[c].resizeable = list->column[c].resizeable;
		col[c].auto_resize = list->column[c].auto_resize;
		col[c].button_passive = list->column[c].button_passive;
	}

	/* inizializza le nuove (se ho aumentato) */
	attributes.wclass = GDK_INPUT_ONLY;
	attributes.event_mask = (GDK_BUTTON_PRESS_MASK |
							 GDK_BUTTON_RELEASE_MASK |
							 GDK_POINTER_MOTION_MASK |
							 GDK_POINTER_MOTION_HINT_MASK |
							 GDK_KEY_PRESS_MASK);
	attributes_mask = GDK_WA_CURSOR;
	attributes.cursor = list->cursor_drag;

	attributes.x =  LIST_WIDTH(list) + 1;
	attributes.y = 0;
	attributes.width = 0;
	attributes.height = 0;

	for (; c < ncols; ++c)
	{
		col[c].title = NULL;
		col[c].area.x = 0;
		col[c].area.y = 0;
		col[c].area.width = 0;
		col[c].area.height = 0;
		col[c].button = NULL;
		col[c].window = NULL;
		col[c].width = 0;
		col[c].min_width = -1;
		col[c].max_width = -1;
		col[c].justification = GTK_JUSTIFY_LEFT;
		col[c].visible = TRUE;
		col[c].width_set = FALSE;
		col[c].resizeable = TRUE; /* TODO: ma questi flags sono indipendenti colonna/colonna? */
		col[c].auto_resize = FALSE;
		col[c].button_passive = FALSE;

		/* TODO: bottoni? credo di no... */
		if (GTK_WIDGET_REALIZED(list))
		{
			col[c].window = gdk_window_new(list->title_window,&attributes,attributes_mask);

			gdk_window_set_user_data(col[c].window,list);
		}
	}

fprintf(stderr,"DEBUG: inizio a liberare le colonne in eccesso\n");

	/* libera le vecchie (se ho diminuito) */
	for (c = l; c < list->columns; ++c)
	{
fprintf(stderr,"DEBUG: libero la colonna %d\n",c); fflush(stderr);

		if (list->column[c].title)
			g_free(list->column[c].title);

		if (list->column[c].button)
		{
fprintf(stderr,"DEBUG: libero il bottone della colonna %d\n",c); fflush(stderr);

			gtk_widget_unparent(list->column[c].button);

			list->column[c].button = NULL;
		}

		/* TODO: finestre? */
	}

fprintf(stderr,"DEBUG: libero l'array delle colonne\n"); fflush(stderr);

	g_free(list->column);

	list->column = col;
	list->columns = ncols;

	/* gli array cells delle righe sono allocati con i memchunks, ma un oggetto memchunk ha a tempo di costruzione codificato dentro il numero di colonne. perci, per poter allocarci roba devo prima ricrearlo con numero di colonne aggiornato, fare il trasferimento, e amen.

	   devo perci ricreare il memchunk
	   riallocarci per le righe esistenti
	   copiarci
	   dellaocare le righe in eccesso
	   liberare quello vecchio
	   riassegnare il nuovo
	*/

	mchunk = g_mem_chunk_new("clist cell mem chunk",sizeof(GtkCell) * ncols,sizeof(GtkCell) * ncols * CLIST_OPTIMUM_SIZE,G_ALLOC_AND_FREE);

fprintf(stderr,"DEBUG: ocols %d, ncols %d, l %d. si %s\n",ocols,ncols,l,l==ncols?"diminuiscono":"aumentano"); fflush(stderr);

	/* modifica le celle */
	if (l == ncols)
		/* si diminuiscono colonne */
		for (ri = list->row_list; ri; ri = g_list_next(ri))
		{
fprintf(stderr,"DEBUG: processo riga %p (dimi)\n",ri); fflush(stderr);

			r = (GtkCListRow *)ri->data;

			cell = g_chunk_new(GtkCell,mchunk);

			for (c = 0; c < l; ++c)
				copy_cell(&cell[c],&r->cell[c]);

			for (; c < ocols; ++c)
			{
				GTK_CLIST_CLASS_FW(list)->set_cell_contents(list,r,c,GTK_CELL_EMPTY,NULL,0,NULL,NULL);

				if (r->cell[c].style)
				{
					if (GTK_WIDGET_REALIZED(list))
						gtk_style_detach(r->cell[c].style);

					gtk_style_unref(r->cell[c].style);
				}
			}
		}
	else
		/* si aumentano colonne */
		for (ri = list->row_list; ri; ri = g_list_next(ri))
		{
fprintf(stderr,"DEBUG: processo riga %p (aume)\n",ri); fflush(stderr);

			r = (GtkCListRow *)ri->data;

			cell = g_chunk_new(GtkCell,mchunk);

			for (c = 0; c < l; ++c)
				copy_cell(&cell[c],&r->cell[c]);

			for (; c < ncols; ++c)
			{
fprintf(stderr,"DEBUG: cella %d\n",c); fflush(stderr);

				cell[c].type = GTK_CELL_EMPTY;
				cell[c].vertical = 0;
				cell[c].horizontal = 0;
				cell[c].style = NULL;
			}

fprintf(stderr,"DEBUG: libero cell_mem_chunk\n"); fflush(stderr);

/*g_mem_chunk_free(list->cell_mem_chunk,r->cell);*/

			r->cell = cell;
		}

	g_mem_chunk_destroy(list->cell_mem_chunk);

	list->cell_mem_chunk = mchunk;

	size_allocate_columns(list,FALSE); /* TODO: TRUE?!? */
	size_allocate_title_buttons(list);

	/* refresh */
	CLIST_REFRESH(list);
}
